@@ -21,6 +21,7 @@
 #include <cstddef>
 #include <cstdint>
 #include <memory>
+#include <set>
 #include <string>
 #include <string_view>
 #include <type_traits>
@@ -78,7 +79,7 @@ struct Var {
     std::string name_;
   };
 };
-typedef std::vector<Var> VarVector;
+using VarVector = std::vector<Var>;
 
 struct Const {
   static constexpr uintptr_t kRefNullBits = ~uintptr_t(0);
@@ -200,7 +201,55 @@ struct Const {
   v128 data_;
   ExpectedNan nan_[4];
 };
-typedef std::vector<Const> ConstVector;
+using ConstVector = std::vector<Const>;
+
+enum class ExpectationType {
+  Values,
+  Either,
+};
+
+class Expectation {
+ public:
+  Expectation() = delete;
+  virtual ~Expectation() = default;
+  ExpectationType type() const { return type_; }
+
+  Location loc;
+
+  ConstVector expected;
+
+ protected:
+  explicit Expectation(ExpectationType type, const Location& loc = Location())
+      : loc(loc), type_(type) {}
+
+ private:
+  ExpectationType type_;
+};
+
+template <ExpectationType TypeEnum>
+class ExpectationMixin : public Expectation {
+ public:
+  static bool classof(const Expectation* expectation) {
+    return expectation->type() == TypeEnum;
+  }
+
+  explicit ExpectationMixin(const Location& loc = Location())
+      : Expectation(TypeEnum, loc) {}
+};
+
+class ValueExpectation : public ExpectationMixin<ExpectationType::Values> {
+ public:
+  explicit ValueExpectation(const Location& loc = Location())
+      : ExpectationMixin<ExpectationType::Values>(loc) {}
+};
+
+struct EitherExpectation : public ExpectationMixin<ExpectationType::Either> {
+ public:
+  explicit EitherExpectation(const Location& loc = Location())
+      : ExpectationMixin<ExpectationType::Either>(loc) {}
+};
+
+typedef std::unique_ptr<Expectation> ExpectationPtr;
 
 struct FuncSignature {
   TypeVector param_types;
@@ -262,6 +311,13 @@ class FuncType : public TypeEntry {
   Type GetResultType(Index index) const { return sig.GetResultType(index); }
 
   FuncSignature sig;
+
+  // The BinaryReaderIR tracks whether a FuncType is the target of a tailcall
+  // (via a return_call_indirect). wasm2c (CWriter) uses this information to
+  // limit its output in some cases.
+  struct {
+    bool tailcall = false;
+  } features_used;
 };
 
 struct Field {
@@ -366,7 +422,9 @@ enum class ExprType {
   TableFill,
   Ternary,
   Throw,
+  ThrowRef,
   Try,
+  TryTable,
   Unary,
   Unreachable,
 
@@ -377,9 +435,9 @@ enum class ExprType {
 const char* GetExprTypeName(ExprType type);
 
 class Expr;
-typedef intrusive_list<Expr> ExprList;
+using ExprList = intrusive_list<Expr>;
 
-typedef FuncDeclaration BlockDeclaration;
+using BlockDeclaration = FuncDeclaration;
 
 struct Block {
   Block() = default;
@@ -402,7 +460,22 @@ struct Catch {
     return var.is_index() && var.index() == kInvalidIndex;
   }
 };
-typedef std::vector<Catch> CatchVector;
+using CatchVector = std::vector<Catch>;
+
+struct TableCatch {
+  explicit TableCatch(const Location& loc = Location()) : loc(loc) {}
+  Location loc;
+  Var tag;
+  Var target;
+  CatchKind kind;
+  bool IsCatchAll() const {
+    return kind == CatchKind::CatchAll || kind == CatchKind::CatchAllRef;
+  }
+  bool IsRef() const {
+    return kind == CatchKind::CatchRef || kind == CatchKind::CatchAllRef;
+  }
+};
+using TryTableVector = std::vector<TableCatch>;
 
 enum class TryKind { Plain, Catch, Delegate };
 
@@ -445,27 +518,28 @@ class MemoryExpr : public ExprMixin<TypeEnum> {
 template <ExprType TypeEnum>
 class MemoryBinaryExpr : public ExprMixin<TypeEnum> {
  public:
-  MemoryBinaryExpr(Var srcmemidx,
-                   Var destmemidx,
+  MemoryBinaryExpr(Var destmemidx,
+                   Var srcmemidx,
                    const Location& loc = Location())
       : ExprMixin<TypeEnum>(loc),
-        srcmemidx(srcmemidx),
-        destmemidx(destmemidx) {}
+        destmemidx(destmemidx),
+        srcmemidx(srcmemidx) {}
 
-  Var srcmemidx;
   Var destmemidx;
+  Var srcmemidx;
 };
 
-typedef ExprMixin<ExprType::Drop> DropExpr;
-typedef ExprMixin<ExprType::Nop> NopExpr;
-typedef ExprMixin<ExprType::Return> ReturnExpr;
-typedef ExprMixin<ExprType::Unreachable> UnreachableExpr;
+using DropExpr = ExprMixin<ExprType::Drop>;
+using NopExpr = ExprMixin<ExprType::Nop>;
+using ReturnExpr = ExprMixin<ExprType::Return>;
+using UnreachableExpr = ExprMixin<ExprType::Unreachable>;
+using ThrowRefExpr = ExprMixin<ExprType::ThrowRef>;
 
-typedef MemoryExpr<ExprType::MemoryGrow> MemoryGrowExpr;
-typedef MemoryExpr<ExprType::MemorySize> MemorySizeExpr;
-typedef MemoryExpr<ExprType::MemoryFill> MemoryFillExpr;
+using MemoryGrowExpr = MemoryExpr<ExprType::MemoryGrow>;
+using MemorySizeExpr = MemoryExpr<ExprType::MemorySize>;
+using MemoryFillExpr = MemoryExpr<ExprType::MemoryFill>;
 
-typedef MemoryBinaryExpr<ExprType::MemoryCopy> MemoryCopyExpr;
+using MemoryCopyExpr = MemoryBinaryExpr<ExprType::MemoryCopy>;
 
 template <ExprType TypeEnum>
 class RefTypeExpr : public ExprMixin<TypeEnum> {
@@ -476,8 +550,8 @@ class RefTypeExpr : public ExprMixin<TypeEnum> {
   Type type;
 };
 
-typedef RefTypeExpr<ExprType::RefNull> RefNullExpr;
-typedef ExprMixin<ExprType::RefIsNull> RefIsNullExpr;
+using RefNullExpr = RefTypeExpr<ExprType::RefNull>;
+using RefIsNullExpr = ExprMixin<ExprType::RefIsNull>;
 
 template <ExprType TypeEnum>
 class OpcodeExpr : public ExprMixin<TypeEnum> {
@@ -488,11 +562,11 @@ class OpcodeExpr : public ExprMixin<TypeEnum> {
   Opcode opcode;
 };
 
-typedef OpcodeExpr<ExprType::Binary> BinaryExpr;
-typedef OpcodeExpr<ExprType::Compare> CompareExpr;
-typedef OpcodeExpr<ExprType::Convert> ConvertExpr;
-typedef OpcodeExpr<ExprType::Unary> UnaryExpr;
-typedef OpcodeExpr<ExprType::Ternary> TernaryExpr;
+using BinaryExpr = OpcodeExpr<ExprType::Binary>;
+using CompareExpr = OpcodeExpr<ExprType::Compare>;
+using ConvertExpr = OpcodeExpr<ExprType::Convert>;
+using UnaryExpr = OpcodeExpr<ExprType::Unary>;
+using TernaryExpr = OpcodeExpr<ExprType::Ternary>;
 
 class SimdLaneOpExpr : public ExprMixin<ExprType::SimdLaneOp> {
  public:
@@ -570,28 +644,28 @@ class MemoryVarExpr : public MemoryExpr<TypeEnum> {
   Var var;
 };
 
-typedef VarExpr<ExprType::Br> BrExpr;
-typedef VarExpr<ExprType::BrIf> BrIfExpr;
-typedef VarExpr<ExprType::Call> CallExpr;
-typedef VarExpr<ExprType::RefFunc> RefFuncExpr;
-typedef VarExpr<ExprType::GlobalGet> GlobalGetExpr;
-typedef VarExpr<ExprType::GlobalSet> GlobalSetExpr;
-typedef VarExpr<ExprType::LocalGet> LocalGetExpr;
-typedef VarExpr<ExprType::LocalSet> LocalSetExpr;
-typedef VarExpr<ExprType::LocalTee> LocalTeeExpr;
-typedef VarExpr<ExprType::ReturnCall> ReturnCallExpr;
-typedef VarExpr<ExprType::Throw> ThrowExpr;
-typedef VarExpr<ExprType::Rethrow> RethrowExpr;
+using BrExpr = VarExpr<ExprType::Br>;
+using BrIfExpr = VarExpr<ExprType::BrIf>;
+using CallExpr = VarExpr<ExprType::Call>;
+using RefFuncExpr = VarExpr<ExprType::RefFunc>;
+using GlobalGetExpr = VarExpr<ExprType::GlobalGet>;
+using GlobalSetExpr = VarExpr<ExprType::GlobalSet>;
+using LocalGetExpr = VarExpr<ExprType::LocalGet>;
+using LocalSetExpr = VarExpr<ExprType::LocalSet>;
+using LocalTeeExpr = VarExpr<ExprType::LocalTee>;
+using ReturnCallExpr = VarExpr<ExprType::ReturnCall>;
+using ThrowExpr = VarExpr<ExprType::Throw>;
+using RethrowExpr = VarExpr<ExprType::Rethrow>;
 
-typedef VarExpr<ExprType::DataDrop> DataDropExpr;
-typedef VarExpr<ExprType::ElemDrop> ElemDropExpr;
-typedef VarExpr<ExprType::TableGet> TableGetExpr;
-typedef VarExpr<ExprType::TableSet> TableSetExpr;
-typedef VarExpr<ExprType::TableGrow> TableGrowExpr;
-typedef VarExpr<ExprType::TableSize> TableSizeExpr;
-typedef VarExpr<ExprType::TableFill> TableFillExpr;
+using DataDropExpr = VarExpr<ExprType::DataDrop>;
+using ElemDropExpr = VarExpr<ExprType::ElemDrop>;
+using TableGetExpr = VarExpr<ExprType::TableGet>;
+using TableSetExpr = VarExpr<ExprType::TableSet>;
+using TableGrowExpr = VarExpr<ExprType::TableGrow>;
+using TableSizeExpr = VarExpr<ExprType::TableSize>;
+using TableFillExpr = VarExpr<ExprType::TableFill>;
 
-typedef MemoryVarExpr<ExprType::MemoryInit> MemoryInitExpr;
+using MemoryInitExpr = MemoryVarExpr<ExprType::MemoryInit>;
 
 class SelectExpr : public ExprMixin<ExprType::Select> {
  public:
@@ -674,8 +748,8 @@ class BlockExprBase : public ExprMixin<TypeEnum> {
   Block block;
 };
 
-typedef BlockExprBase<ExprType::Block> BlockExpr;
-typedef BlockExprBase<ExprType::Loop> LoopExpr;
+using BlockExpr = BlockExprBase<ExprType::Block>;
+using LoopExpr = BlockExprBase<ExprType::Loop>;
 
 class IfExpr : public ExprMixin<ExprType::If> {
  public:
@@ -687,6 +761,15 @@ class IfExpr : public ExprMixin<ExprType::If> {
   Location false_end_loc;
 };
 
+class TryTableExpr : public ExprMixin<ExprType::TryTable> {
+ public:
+  explicit TryTableExpr(const Location& loc = Location())
+      : ExprMixin<ExprType::TryTable>(loc) {}
+
+  Block block;
+  TryTableVector catches;
+};
+
 class TryExpr : public ExprMixin<ExprType::Try> {
  public:
   explicit TryExpr(const Location& loc = Location())
@@ -734,17 +817,17 @@ class LoadStoreExpr : public MemoryExpr<TypeEnum> {
   Address offset;
 };
 
-typedef LoadStoreExpr<ExprType::Load> LoadExpr;
-typedef LoadStoreExpr<ExprType::Store> StoreExpr;
+using LoadExpr = LoadStoreExpr<ExprType::Load>;
+using StoreExpr = LoadStoreExpr<ExprType::Store>;
 
-typedef LoadStoreExpr<ExprType::AtomicLoad> AtomicLoadExpr;
-typedef LoadStoreExpr<ExprType::AtomicStore> AtomicStoreExpr;
-typedef LoadStoreExpr<ExprType::AtomicRmw> AtomicRmwExpr;
-typedef LoadStoreExpr<ExprType::AtomicRmwCmpxchg> AtomicRmwCmpxchgExpr;
-typedef LoadStoreExpr<ExprType::AtomicWait> AtomicWaitExpr;
-typedef LoadStoreExpr<ExprType::AtomicNotify> AtomicNotifyExpr;
-typedef LoadStoreExpr<ExprType::LoadSplat> LoadSplatExpr;
-typedef LoadStoreExpr<ExprType::LoadZero> LoadZeroExpr;
+using AtomicLoadExpr = LoadStoreExpr<ExprType::AtomicLoad>;
+using AtomicStoreExpr = LoadStoreExpr<ExprType::AtomicStore>;
+using AtomicRmwExpr = LoadStoreExpr<ExprType::AtomicRmw>;
+using AtomicRmwCmpxchgExpr = LoadStoreExpr<ExprType::AtomicRmwCmpxchg>;
+using AtomicWaitExpr = LoadStoreExpr<ExprType::AtomicWait>;
+using AtomicNotifyExpr = LoadStoreExpr<ExprType::AtomicNotify>;
+using LoadSplatExpr = LoadStoreExpr<ExprType::LoadSplat>;
+using LoadZeroExpr = LoadStoreExpr<ExprType::LoadZero>;
 
 class AtomicFenceExpr : public ExprMixin<ExprType::AtomicFence> {
  public:
@@ -765,8 +848,8 @@ struct Tag {
 
 class LocalTypes {
  public:
-  typedef std::pair<Type, Index> Decl;
-  typedef std::vector<Decl> Decls;
+  using Decl = std::pair<Type, Index>;
+  using Decls = std::vector<Decl>;
 
   struct const_iterator {
     const_iterator(Decls::const_iterator decl, Index index)
@@ -845,6 +928,13 @@ struct Func {
   BindingHash bindings;
   ExprList exprs;
   Location loc;
+
+  // For a subset of features, the BinaryReaderIR tracks whether they are
+  // actually used by the function. wasm2c (CWriter) uses this information to
+  // limit its output in some cases.
+  struct {
+    bool tailcall = false;
+  } features_used;
 };
 
 struct Global {
@@ -865,7 +955,7 @@ struct Table {
   Type elem_type;
 };
 
-typedef std::vector<ExprList> ExprListVector;
+using ExprListVector = std::vector<ExprList>;
 
 struct ElemSegment {
   explicit ElemSegment(std::string_view name) : name(name) {}
@@ -884,6 +974,7 @@ struct Memory {
 
   std::string name;
   Limits page_limits;
+  uint32_t page_size;
 };
 
 struct DataSegment {
@@ -1001,7 +1092,7 @@ class ModuleField : public intrusive_list_base<ModuleField> {
   ModuleFieldType type_;
 };
 
-typedef intrusive_list<ModuleField> ModuleFieldList;
+using ModuleFieldList = intrusive_list<ModuleField>;
 
 template <ModuleFieldType TypeEnum>
 class ModuleFieldMixin : public ModuleField {
@@ -1116,6 +1207,17 @@ class StartModuleField : public ModuleFieldMixin<ModuleFieldType::Start> {
   Var start;
 };
 
+struct Custom {
+  explicit Custom(const Location& loc = Location(),
+                  std::string_view name = std::string_view(),
+                  std::vector<uint8_t> data = std::vector<uint8_t>())
+      : name(name), data(data), loc(loc) {}
+
+  std::string name;
+  std::vector<uint8_t> data;
+  Location loc;
+};
+
 struct Module {
   Index GetFuncTypeIndex(const Var&) const;
   Index GetFuncTypeIndex(const FuncDeclaration&) const;
@@ -1187,6 +1289,7 @@ struct Module {
   std::vector<Memory*> memories;
   std::vector<DataSegment*> data_segments;
   std::vector<Var*> starts;
+  std::vector<Custom> customs;
 
   BindingHash tag_bindings;
   BindingHash func_bindings;
@@ -1197,6 +1300,21 @@ struct Module {
   BindingHash memory_bindings;
   BindingHash data_segment_bindings;
   BindingHash elem_segment_bindings;
+
+  // For a subset of features, the BinaryReaderIR tracks whether they are
+  // actually used by the module. wasm2c (CWriter) uses this information to
+  // limit its output in some cases.
+  struct {
+    bool simd = false;
+    bool exceptions = false;
+    bool threads = false;
+  } features_used;
+
+  // The BinaryReaderIR tracks function references used by the module, whether
+  // in element segment initializers, global initializers, or functions. wasm2c
+  // needs to emit wrappers for any functions that might get used as function
+  // references, and uses this information to limit its output.
+  std::set<Index> used_func_refs;
 };
 
 enum class ScriptModuleType {
@@ -1249,8 +1367,8 @@ class DataScriptModule : public ScriptModuleMixin<TypeEnum> {
   std::vector<uint8_t> data;
 };
 
-typedef DataScriptModule<ScriptModuleType::Binary> BinaryScriptModule;
-typedef DataScriptModule<ScriptModuleType::Quoted> QuotedScriptModule;
+using BinaryScriptModule = DataScriptModule<ScriptModuleType::Binary>;
+using QuotedScriptModule = DataScriptModule<ScriptModuleType::Quoted>;
 
 enum class ActionType {
   Invoke,
@@ -1276,7 +1394,7 @@ class Action {
   ActionType type_;
 };
 
-typedef std::unique_ptr<Action> ActionPtr;
+using ActionPtr = std::unique_ptr<Action>;
 
 template <ActionType TypeEnum>
 class ActionMixin : public Action {
@@ -1320,7 +1438,7 @@ enum class CommandType {
   First = Module,
   Last = AssertException,
 };
-static const int kCommandTypeCount = WABT_ENUM_COUNT(CommandType);
+constexpr int kCommandTypeCount = WABT_ENUM_COUNT(CommandType);
 
 class Command {
  public:
@@ -1361,7 +1479,7 @@ class ActionCommandBase : public CommandMixin<TypeEnum> {
   ActionPtr action;
 };
 
-typedef ActionCommandBase<CommandType::Action> ActionCommand;
+using ActionCommand = ActionCommandBase<CommandType::Action>;
 
 class RegisterCommand : public CommandMixin<CommandType::Register> {
  public:
@@ -1375,7 +1493,7 @@ class RegisterCommand : public CommandMixin<CommandType::Register> {
 class AssertReturnCommand : public CommandMixin<CommandType::AssertReturn> {
  public:
   ActionPtr action;
-  ConstVector expected;
+  ExpectationPtr expected;
 };
 
 template <CommandType TypeEnum>
@@ -1385,9 +1503,9 @@ class AssertTrapCommandBase : public CommandMixin<TypeEnum> {
   std::string text;
 };
 
-typedef AssertTrapCommandBase<CommandType::AssertTrap> AssertTrapCommand;
-typedef AssertTrapCommandBase<CommandType::AssertExhaustion>
-    AssertExhaustionCommand;
+using AssertTrapCommand = AssertTrapCommandBase<CommandType::AssertTrap>;
+using AssertExhaustionCommand =
+    AssertTrapCommandBase<CommandType::AssertExhaustion>;
 
 template <CommandType TypeEnum>
 class AssertModuleCommand : public CommandMixin<TypeEnum> {
@@ -1396,13 +1514,13 @@ class AssertModuleCommand : public CommandMixin<TypeEnum> {
   std::string text;
 };
 
-typedef AssertModuleCommand<CommandType::AssertMalformed>
-    AssertMalformedCommand;
-typedef AssertModuleCommand<CommandType::AssertInvalid> AssertInvalidCommand;
-typedef AssertModuleCommand<CommandType::AssertUnlinkable>
-    AssertUnlinkableCommand;
-typedef AssertModuleCommand<CommandType::AssertUninstantiable>
-    AssertUninstantiableCommand;
+using AssertMalformedCommand =
+    AssertModuleCommand<CommandType::AssertMalformed>;
+using AssertInvalidCommand = AssertModuleCommand<CommandType::AssertInvalid>;
+using AssertUnlinkableCommand =
+    AssertModuleCommand<CommandType::AssertUnlinkable>;
+using AssertUninstantiableCommand =
+    AssertModuleCommand<CommandType::AssertUninstantiable>;
 
 class AssertExceptionCommand
     : public CommandMixin<CommandType::AssertException> {
@@ -1410,8 +1528,8 @@ class AssertExceptionCommand
   ActionPtr action;
 };
 
-typedef std::unique_ptr<Command> CommandPtr;
-typedef std::vector<CommandPtr> CommandPtrVector;
+using CommandPtr = std::unique_ptr<Command>;
+using CommandPtrVector = std::vector<CommandPtr>;
 
 struct Script {
   WABT_DISALLOW_COPY_AND_ASSIGN(Script);
